from .info import (
    __author__,
    __description__,
    __email__,
    __license__,
    __version__,
)

__all__ = [
    "__author__",
    "__description__",
    "__email__",
    "__license__",
    "__version__",
]
