# import libs

# SECTION: PyThermoDBLink/PyThermoDB
DATASOURCE = "datasource"
EQUATIONSOURCE = "equationsource"


# SECTION: constants
R_J_molK = 8.314462618  # universal gas constant in J/mol.K
T_ref_K = 298.15  # reference temperature in K
P_ref_Pa = 101325.0  # reference pressure in Pa
