# config
from .configs import (
    __version__,
    __author__,
    __description__,
    __email__,
    __license__,
)

__all__ = [
    # config
    "__version__",
    "__author__",
    "__description__",
    "__email__",
    "__license__",
    # app
]
