# import libs

# SECTION: hsg properties
# NOTE: symbols
EnFo_IG_SYMBOL = 'EnFo_IG'  # enthalpy of formation symbol
EnFo_LIQ_SYMBOL = 'EnFo_LIQ'  # enthalpy of formation liquid symbol
EnFo_SOL_SYMBOL = 'EnFo_SOL'  # enthalpy of formation solid symbol
Ent_STD_SYMBOL = 'Ent_STD'  # entropy of formation symbol
GiEnFo_IG_SYMBOL = 'GiEnFo_IG'  # Gibbs free energy of formation symbol
GiEnFo_LIQ_SYMBOL = 'GiEnFo_LIQ'  # Gibbs free energy of formation symbol
GiEnFo_SOL_SYMBOL = 'GiEnFo_SOL'  # Gibbs free energy of formation symbol
En_IG_SYMBOL = 'En_IG'  # ideal gas enthalpy symbol
En_LIQ_SYMBOL = 'En_LIQ'  # liquid enthalpy symbol
En_SOL_SYMBOL = 'En_SOL'  # solid enthalpy symbol
GiEn_IG_SYMBOL = 'GiEn_IG'  # ideal gas Gibbs energy symbol
GiEn_LIQ_SYMBOL = 'GiEn_LIQ'  # liquid Gibbs energy symbol
GiEn_SOL_SYMBOL = 'GiEn_SOL'  # solid Gibbs energy symbol
Cp_IG_SYMBOL = 'Cp_IG'  # ideal gas heat capacity symbol
Cp_LIQ_SYMBOL = 'Cp_LIQ'  # liquid heat capacity symbol
Cp_SOL_SYMBOL = 'Cp_SOL'  # solid heat capacity symbol
VaPr_SYMBOL = 'VaPr'  # vapor pressure symbol
EnVap_SYMBOL = 'EnVap'  # enthalpy of vaporization symbol
EnSub_SYMBOL = 'EnSub'  # enthalpy of sublimation symbol

# NOTE: units
EnFo_IG_UNIT = 'J/mol'  # enthalpy of formation unit
EnFo_LIQ_UNIT = 'J/mol'  # enthalpy of formation liquid unit
EnFo_SOL_UNIT = 'J/mol'  # enthalpy of formation solid unit
Ent_STD_UNIT = 'J/mol.K'  # entropy of formation unit
GiEnFo_IG_UNIT = 'J/mol'  # Gibbs free energy of formation unit
GiEnFo_LIQ_UNIT = 'J/mol'  # Gibbs free energy of formation liquid unit
GiEnFo_SOL_UNIT = 'J/mol'  # Gibbs free energy of formation solid unit
En_IG_UNIT = 'J/mol'  # ideal gas enthalpy unit
En_LIQ_UNIT = 'J/mol'  # liquid enthalpy unit
En_SOL_UNIT = 'J/mol'  # solid enthalpy unit
GiEn_IG_UNIT = 'J/mol'  # ideal gas Gibbs energy unit
GiEn_LIQ_UNIT = 'J/mol'  # liquid Gibbs energy unit
GiEn_SOL_UNIT = 'J/mol'  # solid Gibbs energy unit
Cp_IG_UNIT = 'J/mol.K'  # ideal gas heat capacity unit
Cp_LIQ_UNIT = 'J/mol.K'  # liquid heat capacity unit
Cp_SOL_UNIT = 'J/mol.K'  # solid heat capacity unit
VaPr_UNIT = 'Pa'  # vapor pressure unit
EnVap_UNIT = 'J/mol'  # enthalpy of vaporization unit
EnSub_UNIT = 'J/mol'  # enthalpy of sublimation unit

# NOTE: integration units
Cp_IG_INTEGRATION_UNIT = 'J/mol'  # ideal gas heat capacity integration unit
Cp_LIQ_INTEGRATION_UNIT = 'J/mol'  # liquid heat capacity integration unit
